-------------------------------------------------------------------------------
-- Copyright (c) 2016 Sierra Wireless. All rights reserved.
--
-- Contributors:
--     Randall Pfeifer - initial implementation
--
--	M3DA / DELS Send Handler
--	Sends DELS-formated data to AMM via M3DA protocol. Handles persisting data
--	across reboot.
-------------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- REQUIRES
-- ----------------------------------------------------------------------------

local airvantage = require 'airvantage'
local devicetree = require 'devicetree'
local persist = require 'persist'
local queue_new = require 'queue'
local lock = require 'sched.lock'
local uptime = require 'sched.timer.core'
local racon = require 'racon'
local timer = require 'timer'


-- ----------------------------------------------------------------------------
-- CONSTANTS
-- ----------------------------------------------------------------------------
local EVTID_BASE = 2048					-- base of event id
local EVTID_PERSIST = 'evtid_base'		-- persist to save current evtid base
local EVTQ_PERSIST = 'evts_queued'		-- persist count & duration of queued events

local DEVT_INFO		= 376				-- DELS event to inform AMM of info messages
local DEVT_WARN		= 375				-- DELS event to inform AMM of warning messages
local DEVT_ERROR	= 374				-- DELS event to inform AMM of error messages

local INFORM_SRC	= 1					-- source ofinform events event (1 is ammer app)

local MSCIID_STS_TIME_VALID = "aleos.55010"

local REFTIME = 1512086400				-- 2017-12-01 00:00:00 UTC - time in recent past against
										-- which to compare current time to see if RTC is valid


-- ----------------------------------------------------------------------------
-- Local Functions
-- ----------------------------------------------------------------------------

------------------------------------------------------------------
-- getevtid - get the next event id
-- 
-- Inputs:
--
-- Outputs:
--
-- Returns:
--	event id
--
-- Notes
------------------------------------------------------------------
local function getevtid(self)
	local rtn = self.eventid_base + self.eventid
	self.eventid = self.eventid + 1
	if self.eventid >= EVTID_BASE then
		self.eventid = 0
		self.eventid_base = self.eventid_base + EVTID_BASE
		persist.save(EVTID_PERSIST, self.eventid_base)
	end
	return rtn
end

------------------------------------------------------------------
-- chktime - check if the RTC is valid
-- 
-- Inputs:
--
-- Outputs:
--
-- Returns:
--
-- Notes
--	Call to check if the RTC is valid, usually once at startup and,
--	if clock not correct then, then again when it becomes valid
--	(via devtree notification).
--
--	If time is valid then self.starttime will be nil, and data is
--	Call when the RTC becomes valid (via devtree notification) to
--	correct time stamps of events which were generated before the
--	actual time was known.
------------------------------------------------------------------
local function chktime(self)
	local now = os.time()
	local flushed = false
--	local status = tonumber(devicetree.get(MSCIID_STS_TIME_VALID))
--	if status == 0 then
	-- compare current time to some date in the past to see it clock is valid instead
	-- of relying on MSCIID_STS_TIME_VALID, because there's a small window where where
	-- this flag may not be accurate.
	if now < REFTIME then
		-- Time not valid yet (app starting).
		if not self.starttime then
			self.starttime = now
			log(self.app_name, "WARNING", "RTC invalid - withholding transmission until time is known")
		end
--		log(self.app_name, "WARNING", "---------> RTC invalid: "..now)
	else
		if self.starttime then
			-- Time is now valid and a start time set (time not valid at app start)
			-- Push all of the data with invalid timestamps to AMM, correcting each
			local offset = now - self.starttime
			log(self.app_name, "INFO", "Time updated: sending "..self.tsqueue:len().." delayed events")
			while self.tsqueue:len() > 0 do
				flushed = true
				-- pop event, adjust timestamp, and send it
				local data = self.tsqueue:pop()
				data.timestamp = data.timestamp + offset

--				local msg = "Popped Event: "
--				for k,v in pairs(data) do msg = msg..k.."="..tostring(v)..", " end
--				log(self.app_name, "DEBUG", msg)

				if self:pushRow(data) ~= "ok" then
					log(self.app_name, "WARNING", "->Event not pushed '"..tostring(data).."'")
				end
			end
			self.eqstat[#self.eqstat] = {}
			self.starttime = nil
		end
		-- Check for queued bad-time messages, send Miscellanious info message if there are any.
		-- These are created if GW reboots before time becomes valid, so they're lost. Gotta warn AMM.
		if #self.eqstat[#self.eqstat] > 0 then
			log(self.app_name, "ERROR", "Problem with dropped events list")
		else
			local now = os.time()
			local sent = false
			for i, losses in ipairs(self.eqstat) do
				local count = losses.count
				if count then
					flushed = true
					local span = losses.last - losses.first
					local msg = "Time unavailable: lost "..count.." events spanning "..span.." seconds"
					-- push inform message directly instead of via inform()
					-- to avoid issues with lock/resursion issues calling send() here
					log(self.app_name, "WARNING", msg)
					self:pushRow({	timestamp=now,
												eventid = getevtid(self),
												messageid=DEVT_WARN,
												param0 = INFORM_SRC,
												extra = msg
					})
				end
			end
			if flushed then
				self.eqstat = { {} }
				persist.save(EVTQ_PERSIST, self.eqstat)
			end
		end
		
	end
end

-- ----------------------------------------------------------------------------
-- METHODS
-- ----------------------------------------------------------------------------

------------------------------------------------------------------
-- send - Member func to send the data to AMM
-- 
-- Inputs:
--	data		- table of DELS data parameters to send:
--					messageid
--					param0
--					param1
--					extra
--
-- Outputs:
--
-- Returns:
--
-- Notes
--	Formats M3DA according to AMM requirements. Fills in timestamp,
--	eventid, etc. Some parameters may be omitted, AMM will fill in.
--
--	If system time id not currently valid, the event(s) will be held
--	in a queue until the correct time is known, so that the timestamps
--	can be corrected and the events then sent with the correct time.
--	Call with null data when the time becomes valid to send the time-
--	corrected events immediately instead of waiting for next send.

--
--	Also keep a persisted record of the number of enqueued events and
--	the timestamp of 1st and last event. If a reboot occurs before the
--	time becomes known, it is impossible to re-construct the correct
--	timestamps, but an event can be created from this data to notify
--	of the duration and the number of affected events lost (discarded).
------------------------------------------------------------------
local function m_send(self, data)
    if self.running then
		-- Lock to prevent sending events while dumping the chktime queue
		self.lock:acquire()

		-- Check if there are any events whitheld due to invalid time stamps
		chktime(self)

		if self.asset and self.asset.events and type(data) == "table" then
			if data.timestamp and data.timestamp ~= 0 then
				-- An explicit timestamp specified. If valid, bypass time correction stuff and just send
				local timestamp = tonumber(data.timestamp)
				if timestamp and timestamp > REFTIME then
					data.eventid = getevtid(self)
					if self:pushRow(data) ~= "ok" then
						log(self.app_name, "WARNING", "->Event not pushed '"..tostring(data).."'")
					end
				else
					log(self.app_name, "WARNING", "Invalid explicit timestamp specified for event " .. event.messageid)
				end
			else
				if data.timestamp == 0 then
					-- An event marked with a pre-existing timestamp == 0 adjusts timestamp
					-- to indicate boot time instead of app start time
					local ut = math.floor(uptime.time())
					data.timestamp = os.time() - ut
					log (self.app_name, "DETAIL", "Started " .. os.date("%X", ut) .. " after boot")
				else
					-- timestamp not overridden, use current time
					data.timestamp = os.time()
				end

				data.eventid = getevtid(self)

				if log.musttrace(self.app_name, "DETAIL") then
					local tmp = "Send:"
					for k,v in pairs(data) do tmp = tmp.."  "..tostring(k).."="..tostring(v) end
					log(self.app_name, "DETAIL", tmp)
				end

				if self.starttime then
					-- current time is invalid - enque data so time stamp can be fixed later
					self.tsqueue:push(data)
					if self.tsqueue:len() > self.maxqueue then
						-- Limit queue size by dropping oldest (log it, though)
						local dropped = self.tsqueue:pop()
						local msg = "Too many events queued - dropped: "
						for k,v in pairs(dropped) do msg = msg..k.."="..tostring(v)..", " end
						log(self.app_name, "WARNING", msg)
					end

					if not self.eqstat[#self.eqstat].count then
						self.eqstat[#self.eqstat].count = 1
						self.eqstat[#self.eqstat].first = data.timestamp
					else
						self.eqstat[#self.eqstat].count = self.eqstat[#self.eqstat].count + 1
					end
--[[
					if log.musttrace(self.app_name, "DEBUG") then
						local msg = "Pushed Event: "
						for k,v in pairs(data) do msg = msg..k.."="..tostring(v)..", " end
						log(self.app_name, "DEBUG", msg)
					end
--]]
					self.eqstat[#self.eqstat].last = data.timestamp
					persist.save(EVTQ_PERSIST, self.eqstat)
					log(self.app_name, "DEBUG", "Invalid time: "..self.eqstat[#self.eqstat].count.." events spanning "..self.eqstat[#self.eqstat].last - self.eqstat[#self.eqstat].first.." seconds delayed")
				else
					if self:pushRow(data) ~= "ok" then
						log(self.app_name, "WARNING", "->Event not pushed '"..tostring(data).."'")
					end
--[[
					if log.musttrace(self.app_name, "DEBUG") then
						local msg = "Sent Event: "
						for k,v in pairs(data) do msg = msg..k.."="..tostring(v)..", " end
						log(self.app_name, "DEBUG", msg)
					end
--]]
				end
			end

		elseif data ~= nil then
			log(self.app_name, "WARNING", "->Event not pushed '"..tostring(data).."'")
		end
		self.lock:release()
	end
end

------------------------------------------------------------------
-- pusRow - push row of data to AMM
-- 
-- Inputs:
--	data		- table of DELS data parameters to send:
--					timestamp
--					eventid
--					messageid
--					param0
--					param1
--					extra
--
-- Outputs:
--
-- Returns:
--
-- Notes
--	Replacement self:pushRow for (table:pushRow) is used to provide
--	limited functionality of persisted table:pushData() functionality,
--	but with much less flash wear. Attempts to duplicate alogrithm used
--	by oMG:
--	- Send pushed data into a queue (RAM) and start latency timer (via signal) if not already running.
--	- When latency timer triggers, all available messages are 'peeked' from the queue and sent immediately (synchronous)
--	- If the send was successful then the 'peeked' messages are popped and discarded.
--	- If the send failed M3DA send data is cleared, and re-loaded from queue on next send
--	- Peridically, if there are at least n messages in the queue, the queue is persisted to disk.
--	- Upon orderly shutdown the queue is persisted to disk.
--	- On startup, the queue is restored from persist.
------------------------------------------------------------------
local function m_pushRow(self, data)
	-- set default param values (recommended to always send ALL defined parameters)
	if not data.timestamp then data.timestamp = 0 end
	if not data.eventid then data.eventid = 0 end
	if not data.messageid then data.messageid = 0 end
	if not data.param0 then data.param0 = 0 end
	if not data.param1 then data.param1 = 0 end
	if not data.extra then data.extra = "" end

	-- If max queue size reached send a inform warning. Can infer restoration by next message
	if self.pushqueue:len() == self.maxqueue then
		self.pushqueue:push {	messageid = DEVT_WARN,
								param0 = INFORM_SRC,
								param1 = 0,
								timestamp = os.time(),
								eventid = getevtid(self),
								extra = "Max Tx queue size of " .. self.maxqueue .. " reached" }
		log(self.app_name, "WARNING", "Max Tx queue size of %d reached", self.maxqueue)
	end
	
	if self.pushqueue:len() < self.maxqueue then
		self.pushqueue:push(data)
	end
	sched.signal(self.app_name, "SEND")

	return "ok"
end

------------------------------------------------------------------
-- inform - Inform AMM of internal conditions/errors
-- 
-- Inputs:
--	status	- "warning", "error", or "info"
--	fmt		- string format that holds the log text the same way as string.format does
--	varargs	- additional arguments can be provided (as with string.format)
--
-- Outputs:
--
-- Returns:
--
-- Notes
------------------------------------------------------------------
local function m_inform(self, status, fmt, ...)
	local msg = string.format(fmt, ...)
	local evts = {info=DEVT_INFO, warning=DEVT_WARN, error=DEVT_ERROR}
	
	if evts[status] then
		self:send({messageid=evts[status], param0 = INFORM_SRC, extra=msg})
		log(self.app_name, "INFO", "Inform AMM %s: %s", status, msg)
	else
		log(self.app_name, "WARNING", "Unrecognized inform status %s: %s", status, msg)
	end
end

-- Timer callback to peridically save the Tx queue in case of unexpected power loss.
-- Only save if events are past threshold to prevent minimal unsent data from creating
-- excessive saves. Note that queue save has built-in checks to eleminate redundant writes, too.
local function savequeue(self)
	local size = self.pushqueue:len()
	if (size == 0 or size >= self.persistthresh) and self.pushqueue:save(self.persistname, false) then
		log(self.app_name, "DETAIL", "Persisted %d records to M3DA Tx queue", self.pushqueue:len())
	else
		log(self.app_name, "DEBUG", "queue not saved")
	end
end

------------------------------------------------------------------
-- run - Start M3DA send operations
-- 
-- Inputs:
--
-- Outputs:
--
-- Returns:
--
-- Notes
--	Start M3DA send operations coroutine.
--	When triggered by "SEND" signal, wait for latency time, and then
--	send all available data from the Tx queue and trigger the M3DA
--	send policy to send it. If data is successfully sent, remove it
--	from the Tx queue.
------------------------------------------------------------------
local function m_run(self)
	if self.running then
		log(self.app_name, "WARNING", "Attempted to start already-running M3DA task")
		return
	end

    self.running = true
	
	self.pushqueue:restore(self.persistname)
	log(self.app_name, "INFO", "Restored %d Persisted records to M3DA Tx queue", self.pushqueue:len())

	if self.persistrate > 0 then
		self.persisttimer = timer.periodic(self.persistrate, savequeue, self)
	end
	
	-- Init timestamp check
	self:send(nil)

	self.connect_task = sched.run (
        function()
            repeat
                local signal = sched.wait(self.app_name, "*")
				log(self.app_name, "DEBUG", "m3da executing on signal " .. signal)
				if signal == "SEND" then
					sched.wait(self.latency)
					repeat
						local data = self.pushqueue:get(self.maxsend)
						log(self.app_name, "DETAIL", "Pushing %d / %d Records ...", #data, self.pushqueue:len())
						for i, event in ipairs(data) do
							self.asset.events:pushRow(event)
						end
						local success, err = racon.triggerPolicy(self.app_name)
						if success == "ok" then
							log(self.app_name, "DEBUG", "Send success - drop " .. #data .. " records")
							for i = 1, #data do self.pushqueue:pop() end
						else
							log(self.app_name, "DETAIL", "M3DA Send failure (" .. #data .. " events) - " .. err)
							self.asset.events:reset()	-- dump unsent data - will be reloaded next try
						end
					until self.pushqueue:len() == 0 or not self.running
				elseif signal == "SAVE" then
					-- persist queue
				elseif signal == "STOP" then
					self.running = false
				else
					log(self.app_name, "DEBUG", "m3da no action on signal " .. signal)
				end
				log(self.app_name, "DEBUG", "m3da action complete")
            until not self.running
			self:send("Stopped")
			log(self.app_name, "INFO", "m3da task exiting ...")
        end
    )
	sched.wait()	-- yield to start thread
end

------------------------------------------------------------------
-- stop - Stop M3DA send operations
-- 
-- Inputs:
--
-- Outputs:
--
-- Returns:
--
-- Notes
--	Stops all M3DA send processes ands immediately persists the
--	Tx queue to flash.
------------------------------------------------------------------
local function m_stop (self)
	if self.running then
		self.running = false
		sched.signal(self.app_name, "STOP")
		if self.persisttimer then
			timer.cancel(self.persisttimer)
			self.persisttimer = nil
		end
		-- Force queue persist on stop. Do here in case triggerPolicy is
		-- blocked trying to send.
		log(self.app_name, "INFO", "Persisting %d records in M3DA Tx queue", self.pushqueue:len())
		self.pushqueue:save(self.persistname, true)
	end
end

------------------------------------------------------------------
-- configure - Configure M3DA parameters
-- 
-- Inputs:
--	config			- table of parameter / value pairs
--
-- Outputs:
--
-- Returns:
--
-- Notes
------------------------------------------------------------------
local function m_configure(self, config)
	-- Maximum queue size
	if tonumber(config.maxqueue) then
		self.maxqueue = tonumber(config.maxqueue)
	end
	
	-- Maximum events in a message
	if tonumber(config.maxsend) then
		self.maxsend = tonumber(config.maxsend)
	end
	
	-- latency is the send latency in seconds. Items sent within this period will
	-- be sent with the one initiating the send.
	if tonumber(config.latency) then
		self.latency = tonumber(config.latency)
	end
	
	-- set persist rate. 0 cancels automatic persist and triggers persist now
	if tonumber(config.persistrate) then
		self.persistrate = tonumber(config.persistrate)
		
		if self.persisttimer then
			timer.cancel(self.persisttimer)
			self.persisttimer = nil
		end
		if self.running then
			if tonumber(config.persistrate) > 0 then
				self.persisttimer = timer.periodic(tonumber(config.persistrate), savequeue, self)
			else
				savequeue(self)
			end
		end
	end
	
	-- persistthresh sets number of queued items that need to be present to save
	if tonumber(config.persistthresh) then
		self.persistthresh = tonumber(config.persistthresh)
	end
	
	-- persistname is name of persist to save queue
	if config.persistname then
		self.persistname = self.app_name .. tostring(config.persistname)
	end

	-- queuetail is size of queue debug tail (retained stale data)
	if config.queuetrail then
		self.pushqueue:settrail(tonumber(config.queuetrail))
	end
end

------------------------------------------------------------------
-- isconnected - Dummy Member func to keep other modules happy
-- 
-- Inputs:
--
-- Outputs:
--
-- Returns:
--
-- Notes
------------------------------------------------------------------
local function m_isconnected(self)
    return true, "connected"
end

------------------------------------------------------------------
-- create - Create an instance of m3da
-- 
-- Inputs:
--	app_name		- Name of the app (for logging etc)
--	asset_name		- Name of the asset to hold send table
--	path			- Data path of M3DA data
--
-- Outputs:
--
-- Returns:
--	m3da instance	- sucess
--	nil, error		- failure
--
-- Notes
------------------------------------------------------------------
local function m_create(app_name, asset_name, path)
    local m = {}
	local err

	airvantage.init()
	devicetree.init()
	racon.init()

    m.app_name = app_name
    m.configure = m_configure
    m.run = m_run
    m.stop = m_stop
    m.isconnected = m_isconnected
    m.send = m_send
	m.inform = m_inform
	m.pushRow = m_pushRow
	
	m.tsqueue = queue_new()
	
	-- persisted table to hold list of stats describing occurrance of
	-- lost events due to lack of known time prior to reboot.
	-- Upon startup, last stat in the list is initially empty (either
	-- newly created if existing stats or re-used if last is empty).
	-- This last stat will be discarded when the respective queued events
	-- are time-adjusted and sent, or sent after next boot as warning message.
	m.eqstat = persist.load(EVTQ_PERSIST)
	if m.eqstat == nil or #m.eqstat < 1  then
		m.eqstat = { {} }
		persist.save(EVTQ_PERSIST, m.eqstat)
		log(app_name, "INFO", "EVTQ_PERSIST initialized")
	end
	-- Ensure there's a fresh instance slot on top at boot.
	if m.eqstat[#m.eqstat].count then
		table.insert(m.eqstat, {})
		persist.save(EVTQ_PERSIST, m.eqstat)
		log(app_name, "INFO", "EVTQ_PERSIST new tip added")
	end

    m.running = false
	
	m.lock = lock:new()
	
	-- Upper part of event id is persisted and incremented on boot
	-- or when the lower part exceeds the base. Guarantees unique
	-- event id for every message without too much flash wear.
	m.eventid_base = persist.load(EVTID_PERSIST)
	if type(m.eventid_base) ~= 'number' then
		m.eventid_base = 0	-- ToDo: make event id persistent
	else
		m.eventid_base = m.eventid_base + EVTID_BASE
	end
	persist.save(EVTID_PERSIST, m.eventid_base)
	m.eventid = 0

	-- Make sure send policy exists. Named after app_name
	local policy = 'config.data.policy.' .. app_name
	local pn, pd = devicetree.get(policy)
	if not pd then
		log(app_name, "INFO", "Creating send policy " .. policy)
		devicetree.set(policy .. ".manual", true)
	end

	-- table for m3da transfer (Tx queue)
	m.pushqueue = queue_new()

	-- Default m3da send parameters
	m.maxqueue = 10000
	m.persistthresh = 1
	m.persistrate = 0
	m.latency = 5
	m.maxsend = 0
	m.persistname = app_name .. "_send_queue"
	
	m.asset, err = airvantage.newAsset(asset_name)
	if not m.asset then
		return nil, err
	end
	m.asset:start()
	m.asset.events = m.asset:newTable(	path,
										{"timestamp", "eventid", "messageid", "param0", "param1", "extra"},
										"ram",
										app_name,
										true)

	m.reftime = REFTIME
    return m
end

return m_create

