-------------------------------------------------------------------------------
-- Copyright (c) 2015 Sierra Wireless. All rights reserved.
--
-- Contributors:
--     Randall Pfeifer - initial implementation
--
--	Main application module.
-------------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- REQUIRES
-- ----------------------------------------------------------------------------

local sched         = require "sched"
local log           = require "log"
local devicetree    = require "devicetree"
local airvantage    = require "airvantage"
local persist       = require "persist"
local timer         = require "timer"
local serial        = require 'serial'
local systime       = require "sched.timer.core"
local os            = require "os"
local psignal       = require "posixsignal"
local string        = require "string"
local bit           = require "bit32"
local pkginfo       = require "pkginfo"

-- application specific code module constructors
local _bbem = require 'bbem'
local _BnB = require 'BnB'
local _m3da = require 'm3da'
local _telemetry = require 'telemetry'

-- Primary code objects
local transport
local BnB
local telemetry
local wanstate = require 'wanstate'

-- ----------------------------------------------------------------------------
-- CONSTANTS
-- ----------------------------------------------------------------------------

-- Set to false to disable debugging interface --
local DEBUG = true

local APP_NAME = "AMMER"
local DEVICE_ID = ""
local CONFIG_ASSET_NAME = "ammer_config"

-- ----------------------------------------------------------------------------
-- VARIABLES
-- ----------------------------------------------------------------------------

-- These are set to point to either ignition input or I/O input in device tree
local IGNITION_SENSE_LINE
local IGNITION_SENSE_TIME

-- validators for common parameters
local function validate_posnum(pname, pvalue)
	if tonumber(pvalue) and tonumber(pvalue) >= 0 then
		return true
	end
	return false
end

-- system options default values, validator (mandatory, return true if valid), and optional action handlers for changed setting
local default_system_options = {
	sys_log_level = {
		value = "INFO",
		validate =
			function(pname, pvalue)
				local valid = {NONE=true,ERROR=true,WARNING=true,INFO=true,DETAIL=true,DEBUG=true,ALL=true}

				return valid[pvalue]
			end,
		handler =
			function(pname, pvalue)
				log.setlevel(pvalue, "AMMER")
			end
	},
	sys_tx_persist_rate = {
		value = 30,
		validate = validate_posnum,
		handler =
			function(pname, pvalue)
				transport:configure {persistrate = pvalue}
			end
	},
	sys_tx_persist_threshold = {
		value = 5,
		validate = validate_posnum,
		handler =
			function(pname, pvalue)
				transport:configure {persistthresh = pvalue}
			end
	},
	sys_tx_latency = {
		value = 5,
		validate = validate_posnum,
		handler =
			function(pname, pvalue)
				transport:configure {latency = pvalue}
			end
	},
	sys_tx_maxsend = {
		value = 256,
		validate = validate_posnum,
		handler =
			function(pname, pvalue)
				transport:configure {maxsend = pvalue}
			end
	},
	sys_tx_maxqueue = {
		value = 10000,
		validate = validate_posnum,
		handler =
			function(pname, pvalue)
				transport:configure {maxqueue = pvalue}
			end
	},
	sys_tx_queuetrail = {
		value = 500,
		validate = validate_posnum,
		handler =
			function(pname, pvalue)
				transport:configure {queuetrail = pvalue}
			end
	}
}

-- Empty file created to indicate device boot has been processed
local bootflagfile = "/tmp/RA/ammerbootflag"

-- Table of used MSCI parameter definitions (from MSCIIDs.h)
local msciid = {
	STS_POWERIN						= "aleos.266",
	STS_MODEM_SENT_PERSIST			= "aleos.10283",
	STS_MODEM_RECV_PERSIST			= "aleos.10284",
	CFG_LOW_PWR_IGNITION_TIMER		= "aleos.11131",
	CFG_LOW_PWR_IO_LINE_TIMER		= "aleos.11143",
	STS_TIME_VALID					= "aleos.55010",
	MP70_VEH_BEHAVE_EN				= "aleos.10995",
	STS_DRECKON_SENSOR_CALIBRATED	= "aleos.10996",
	CFG_TELEM_CANBUS_ENABLED		= "aleos.59077",
	CFG_TELEM_IGN_SOURCE			= "aleos.59076",
	GROUP_CANBUS					= "aleos.10312",
}

-- DELS event map
local events  = {
	telemetry                  = {messageid = 0,   name = "Telemetry Data"},
	appinfo                    = {messageid = 1,   name = "App Info"},
	caninfo                    = {messageid = 1,   name = "CANbus availability"},
	startup                    = {messageid = 256, name = "Gateway Start Up"},
	heartbeat                  = {messageid = 270, name = "Gateway Heartbeat"},
	uncleanshutdown            = {messageid = 269, name = "Gateway Unclean Shutdown"},
	shutdown                   = {messageid = 257, name = "Gateway Shut Down"},
	wanlinktxbytes             = {messageid = 520, name = "WAN Link Tx Bytes"},
	wanlinkrxbytes             = {messageid = 521, name = "WAN Link Rx Bytes"},
	watchdogreboot             = {messageid = 278, name = "Gateway Software Watchdog System Reboot Determination"},
	gpslocation                = {messageid = 772, name = "Gateway Location Via Internal GPS"},
	gpsspeed                   = {messageid = 773, name = "Gateway Speed Via Internal GPS"},
	gpsfix                     = {messageid = 774, name = "Gateway Internal GPS Fix Status"},
	ignitionstate              = {messageid = 775, name = "Ignition State"},
	bootignitionstate          = {messageid = 775, name = "Boot Ignition State"},
	acceleration_incident_start = {messageid = 70144, name = "Vehicle Behaviour Acceleration Incident Start"},
	acceleration_incident_end  = {messageid = 70145, name = "Vehicle Behaviour Acceleration Incident End"},
	deceleration_incident_start = {messageid = 70146, name = "Vehicle Behaviour Deceleration Incident Start"},
	deceleration_incident_end  = {messageid = 70147, name = "Vehicle Behaviour Deceleration Incident End"},
	cornering_incident_start   = {messageid = 70148, name = "Vehicle Behaviour Cornering Incident Start"},
	cornering_incident_end     = {messageid = 70149, name = "Vehicle Behaviour Cornering Incident End"},
--	crash_incident_start       = {messageid = 70150, name = "Vehicle Behaviour Crash Incident Start"},
--	crash_incident_end         = {messageid = 70151, name = "Vehicle Behaviour Crash Incident End"},
	dreckon_sensor_calibration  = {messageid = 70152, name = "Vehicle Behaviour Sensor calibration"}
}

-- Table to relate telemetry parameter by name to DELS ID and format
local DELSparams = {
	["vehicle_speed"]					= {id =  13, format = "I"},		-- MPH (int)
	["engine_speed"]					= {id =  12, format = "I"},		-- RPM (int)
	["throttle_position"]				= {id =  17, format = "I"},		-- % (int)
	["odometer"]						= {id =  97, format = "I"},		-- Miles (int)
	["fuel_level"]						= {id =  47, format = "I"},		-- % (int)
	["engine_coolant_temp"]				= {id =   5, format = "I"},		-- Deg F (int)
	["mil_status"]						= {id = 317, format = "I"},		-- 0 (off) | 1 (on)
	["battery_voltage"]					= {id =  83, format = "S"},		-- Volts (1 dec)
	["pto_status"]						= {id =  30, format = "I"},		-- 0 (off) | 1 (on)
	["seatbelt_fastened"]				= {id = 193, format = "I"},		-- 0 (fastened) | 1 (not fastened)
	["brake_switch_status"]				= {id = 341, format = "I"},		-- 0 (off) | 1 (on)
	["ambient_air_temperature"]			= {id =  70, format = "S"},		-- Deg C (1 dec)
	["trip_fuel_consumption"]			= {id = 201, format = "I"},		-- Gallons (int)
	["distance_since_dtc_cleared"]		= {id =  49, format = "I"},		-- Miles (int)
	["transmission_fluid_temperature"]	= {id = 145, format = "S"},		-- Deg F (1 dec)
	["engine_run_time"]					= {id =  31, format = "I"},		-- Seconds (int)
	["vin"]								= {id =   2, format = "S"},		-- string
	["dtcs"]							= {id =   1, format = "S"}		-- comma-separated strings
}
-- ----------------------------------------------------------------------------
-- HELPER FUNCTIONS
-- ----------------------------------------------------------------------------

-- encode GPS coord into 9.23 format
local function encodeGps(coord)
	-- This nasty little hack is required because bit32 doesn't handle 32 bits
	if coord < 0 then
		-- Remember negate is complement/increment (hence the "1" and MSB)
		return (bit.bxor(math.floor((-coord) * (2 ^ 23) + 0.5), 0x7fffffff) + 0x80000001)
	else
		return (math.floor(coord * (2 ^ 23) + 0.5))
	end
end

-- Debugging routine to display bytes in a string
function showdata(str)
	if type(str) ~= "string" then
		return ""
	end

	local s = ""
	local i = 1
	while(string.byte(str,i)) do
		s = s..(string.format("%02x ",string.byte(str,i)))
		i = i + 1
	end
	return s
end

------------------------------------------------------------------
-- recieve_config - AirVantage data handler callback to recieve data 
-- 
-- Inputs:
-- 	assetInstance	- Asset Instance
-- 	value			- Value of recieved data
-- 	path			- data model path
--
-- Outputs:
-- 	telemetry.options
--
-- Returns:
-- 	"ok" to indicate to AirVantage that data has been handled
--
-- Notes
--	Recieves configuration data from AirVantage system and saves to
--	configuration parameters in telemetry opbject. Updates persistent
--	object store with new data.
------------------------------------------------------------------
local function recieve_config(assetInstance, value, path)
    --print("*** Recieved: ", assetInstance, value, path)
    local err = nil
    for pname,pvalue in pairs(value) do 

    	if type(pvalue) ~= "string" then
    		-- Sometimes we get wrong data type (function) despite specifying string. Fix it.
    		pvalue = ""
    	end
       	log(APP_NAME, "INFO", 'Configure %s="%s"', pname, tostring(pvalue))

		-- process as a system option if it's in defaults table
		-- system option value are all strings
		if default_system_options[pname] then
			if not default_system_options[pname].validate(pname, pvalue) then
				-- empty or invalid value specified - use default
				log(APP_NAME, "WARNING", "Invalid value for system option %s: %s", pname, tostring(pvalue))
				telemetry.options[pname] = default_system_options[pname].value
			else
				telemetry.options[pname] = pvalue
			end
			if default_system_options[pname].handler then
				default_system_options[pname].handler(pname, telemetry.options[pname])
			end
		else
			err = BnB:set_option(telemetry.options, pname, pvalue)
			if err ~= "ok" then
				log(APP_NAME, "ERROR", "Configure error (%s) %s=%s", err, pname, tostring(pvalue))
				-- return err
			end
		end
    end

	persist.save(CONFIG_ASSET_NAME, telemetry.options)
    telemetry:reconfigure()
    return "ok"
end

-- Get configuration as a table of strings, suitable for
-- sending to AMM (table of value=setting).
local function get_config(assetInstance, value, path)
	local options = {}
	local pnames = {
		{ext="enable", int="en"},
		{ext="threshold", int="th"},
		{ext="min", int="mn"},
		{ext="max", int="mx"}
	}

	for k,v in pairs(telemetry.options) do
		if default_system_options[k] then
			-- if param is in system defaults then it's a (simple) system setting
			options[k] = tostring(v)
		else
			-- it's a compound telemetry setting (enable, threshold, min, max)
			local PIDInfo = telemetry.streamer.PIDInfoByName[k]
			if PIDInfo then	-- make sure it's a valid option
				local opts = ""
				for i = 1, #pnames do
					if not PIDInfo.mask or not PIDInfo.mask[pnames[i].int] then
						-- hide masked sub-params according to master table
						opt = pnames[i].ext .. '=' .. v[pnames[i].int]
						if opts == "" then
							opts = opt
						else
							opts = opts .. "," .. opt
						end
					end
				end
				options[k] = opts
			end
		end
	end
	return options
end

-- airvantage handler callback to send the entire configuration to AMM.
-- Ignores any incoming parameters.
local function request_config(assetInstance, value, path)
	log(APP_NAME, "INFO", "request_config: path=" .. tostring(path) .. " value=" .. tostring(value))
	local options = get_config(assetInstance, value, path)
	assert(assetInstance:pushData("uplink", options))
	return "ok"
end

-- ----------------------------------------------------------------------------
-- DELS Event Generation Functions (from eventreporter app)
-- ----------------------------------------------------------------------------
-- Table to associate MP70 accelerometer incidents with dev tree params
-- start events and end events share structure, so group them
local vbistartparms = {
	acceleration_incident_start = "system.vehicle.accel.event_start",
	deceleration_incident_start = "system.vehicle.brake.event_start",
	cornering_incident_start = "system.vehicle.side.event_start"
--	crash_incident_start = "system.vehicle.crash.event_start"
}
local vbiendparms = {
	acceleration_incident_end = "system.vehicle.accel.event_end",
	deceleration_incident_end = "system.vehicle.brake.event_end",
	cornering_incident_end = "system.vehicle.side.event_end"
--	crash_incident_end = "system.vehicle.crash.event_end"
}

--[[
-- Until params are available by name ...
-- (delete these when MP70 is updated)
local vbistartparms = {
	acceleration_incident_start = "aleos.10985",
	deceleration_incident_start = "aleos.10987",
	ing_incident_start = "aleos.10989",
	crash_incident_start = "aleos.10991"
}
local vbiendparms = {
	acceleration_incident_end = "aleos.10986",
	deceleration_incident_end = "aleos.10988",
	cornering_incident_end = "aleos.10990",
	crash_incident_end = "aleos.10992"
}
--]]

-- Split a (comma) separated string into an array of fields
local function split(inputstr, sep)
	if sep == nil then
		sep = ","
	end
	local t={} ; i=1
	for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
		t[i] = str
		i = i + 1
	end
	return t
end

-- Collect data for event t and generate an event from it. Some events
-- also apply the telemetry config filter to suppress unwanted events.
-- Some events also use v parameter, which contains the devicetree data
-- that triggered the event. Not all events use this (though more probably
-- should). Note that calling devicetree.get() can cause AAF to yield,
-- possibly causing unexpected results.
local function getevent(t, v)
    local event = {}
	event.messageid = events[t].messageid
	if t == "startup" then
		event.timestamp = 0		-- flag to timestamp as initial boot time
		event.param0 = 1		-- boot cause (not available, so use ignition)
		event.param1 = (tonumber(devicetree.get(msciid.STS_POWERIN)) or 0) * 1000
		event.extra = "ESN="..devicetree.get("system.hw_info.global_id")..
						",platform="..devicetree.get("system.hw_info.product_str") ..
						",board="..devicetree.get("system.hw_info.hw_ver")

	elseif t == "appinfo" then
		event.extra = "AMMER Ver = " .. pkginfo.manifest.components[1].version ..
				      ", build = " .. pkginfo.build ..
				      ", commit = " .. pkginfo.commit
		log(APP_NAME, "INFO", event.extra)

	elseif t == "caninfo" then
		if devicetree.get(msciid.GROUP_CANBUS) == "1" then
			if devicetree.get(msciid.CFG_TELEM_CANBUS_ENABLED) == "1" then
				event.extra = "CAN=Enabled"
			else
				event.extra = "CAN=Disabled"
			end
		else
			event.extra = "CAN=N/A"
		end
		log(APP_NAME, "INFO", event.extra)

	elseif t == "ignitionstate" or t == "bootignitionstate" then
		if t == "bootignitionstate" then
			event.timestamp = 0		-- flag to timestamp as initial boot time
		end

		local ignstate = devicetree.get(IGNITION_SENSE_LINE)
		event.param0 = ignstate
		event.param1 = tonumber(devicetree.get(IGNITION_SENSE_TIME))
		
		local tmp = {["ignition_state"] = ignstate}
		telemetry:filter(tmp)
		if tmp["ignition_state"] == nil then
			event = nil
		end

	elseif t == "heartbeat" then
		local tmp = { [t] = 0 } -- no threshold to filter on
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		else
			persist.save("lastheartbeat", os.time() * 1000)
			event.param0 = math.floor(systime.time())
		end

	elseif t == "uncleanshutdown" then
		local lasthb = persist.load("lastheartbeat")
		if type(lasthb) == "number" then
			-- work-around for brain-dead bit32 functions
			event.param0 = math.floor(lasthb / 0x100000000)
			event.param1 = lasthb - (event["param0"] * 0x100000000)
		end

	elseif t == "shutdown" then
		-- shutdown info persisted on last shutdown, because there's no direct way to discern
		-- app shutdown from OS shutdown. So, we catch the last one before actual reboot
		-- (want to show os shutdown, not app)
		local lastsd = persist.load("lastshutdown")
		if lastsd then
			if lastsd.time >= transport.reftime then
				-- use shutdown timestamp if valid
				event.timestamp = lastsd.time
			end
			event.param0 = lastsd.reason
		else
			-- no shutdown info - no event
			event = nil
		end
	elseif t == "gpsfix" then
		local qmap = {[0]=0x3000, [1]=0x0000, [2]=0x1000}
	    event.param0 = devicetree.get('system.gps.fix')
		-- Pack the status word
		-- assume no antenna fault
		-- assume 2D fix (not supported)
		-- assume PDOPC Ok (not supported)
		event.param1 = devicetree.get('system.gps.sat_cnt') +							-- NSIV
						  (math.min(3, devicetree.get('system.gps.sat_cnt')) * 64) +	-- NSU
						  (qmap[devicetree.get('system.gps.qual')] or qmap[0])			-- FD

-- The remaining parameters are filtered via the telemetry filtering mechanism
-- so that the reporting frequency may be configured. The data for these is
-- generally polled rathar than triggered on change.

	elseif t == "wanlinktxbytes" then
		event = wanstate.getCellTxEvent()
		
		if event then
			local tmp = { [t] = event.param1 }
			telemetry:filter(tmp)
			if tmp[t] == nil then
				event = nil
			end
		end

	elseif t == "wanlinkrxbytes" then
		event = wanstate.getCellRxEvent()
		
		if event then
			local tmp = { [t] = event.param1 }
			telemetry:filter(tmp)
			if tmp[t] == nil then
				event = nil
			end
		end

	elseif t == "gpslocation" then
		-- Read GPS coords, convert to 9.23 format
		local gpsdata = devicetree.get({"system.gps.latitude", "system.gps.longitude"})
		event.param0 = encodeGps(gpsdata['system.gps.latitude'])
		event.param1 = encodeGps(gpsdata['system.gps.longitude'])

		-- Run thru filter to reduce data usage
		local tmp = { [t] = {latitude = gpsdata['system.gps.latitude'], longitude = gpsdata['system.gps.longitude']} }
		telemetry:filter(tmp)
		if tmp[t] == nil or (gpsdata['system.gps.latitude'] == 0 and gpsdata['system.gps.longitude'] == 0) then
			event = nil
		end

	elseif t == "gpsspeed" then
		-- Convert to M/H, filter via telemetry filter
	    event.param0 = devicetree.get('system.gps.speed') * 1000
		
		local tmp = { [t] = event.param0 }
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		end
	-- set of accelerometer start events
	elseif vbistartparms[t] then
		-- Split the event data string into individual parameters
		-- Use the data from the event - presumably this is queued so
		-- we won't miss events.
		local params = split(v)
		event.param0 = tonumber(params[1])	-- 1st reading past threshold
		event.param1 = tonumber(params[2])	-- event counter
		-- Run thru telmetry filter
		-- Note that ext threshold (always triggers) is used, and
		-- min rate will never be respected, because this is not polled.
		local tmp = { [t] = 0 }
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		end
	-- set of accelerometer end events
	elseif vbiendparms[t] then
		-- Split the event data string into individual parameters
		-- Use the data from the event - presumably this is queued so
		-- we won't miss events.
		local params = split(v)
		event.param0 = tonumber(params[2])	-- peak accel
		event.param1 = tonumber(params[4])	-- event counter
		event.extra = "duration=" .. params[1] .. ",average=" .. params[3]

		-- Run thru telmetry filter
		-- Note that ext threshold (always triggers) is used, and
		-- min rate will never be respected, because this is not polled.
		local tmp = { [t] = 0 }
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		end
	elseif t == "dreckon_sensor_calibration" then
		event.param0 = tonumber(devicetree.get(msciid.STS_DRECKON_SENSOR_CALIBRATED))
		local tmp = { [t] = 0 }
		telemetry:filter(tmp)
		if tmp[t] == nil or devicetree.get(msciid.MP70_VEH_BEHAVE_EN) ~= "1" then
			event = nil
		end
	else
		--unhandled event
		event = nil
	end

	return event
end

local function generateevent(t, v)
    local content = getevent(t, v)
	if content then
		transport:send(content)
	end
end
local function msciidtoevent(vars)
	for k,v in pairs(vars) do
		log(APP_NAME, "DEBUG", "msciidtoevent: "..k.."="..v)
	end
    if vars[msciid.STS_TIME_VALID] then
		transport:send(nil)
	end
    if vars["system.gps.fix"] or
		vars["system.gps.qual"] or
	   (vars["system.gps.sat_cnt"] and vars["system.gps.sat_cnt"] <= 4) then
    	generateevent("gpsfix")
    end
    if vars["system.gps.speed"] then
    	generateevent("gpsspeed")
    end
	if vars[msciid.STS_DRECKON_SENSOR_CALIBRATED] then
		generateevent("dreckon_sensor_calibration")
	end

	-- check list of vehicle behaviour events
	-- Check for start events first to ensure that start
	-- always preceeds stop (when they appear simultanious)
	for evt,dtname in pairs(vbistartparms) do
		if vars[dtname] then
			generateevent(evt, vars[dtname])
		end
	end

	for evt,dtname in pairs(vbiendparms) do
		if vars[dtname] then
			generateevent(evt, vars[dtname])
		end
	end

    return "ok"
end

-- Send an update of WAN conditions to AMM. This is called
-- when a change in network status is detected. WAN link state
-- generation always looks at entire WAN state, not just changes.
local WanUpdateFilter = {}
local function updateWanEvents()
	local links = wanstate.extractWanLinks()
	local events = wanstate.createWanLinkEvents(links)
	for i, event in ipairs(events) do
		-- Filter out redundant events.
		-- Some events are complementary (WAN up/down and active/notactive),
		-- so the efid field is added to the event when generated; these pairs
		-- have the same efid so that they are compared together instead of as
		-- separate events. The efid essentially creates a slot for each event
		-- to compare that even's previous state to send only on transtition.
		-- Some events add the DPI to the efid, so that the same event applied
		-- to different interface will effectively create a unique slot for that
		-- event/DPI combination. The extra field is not checked, because no
		-- events use that parameter.
		if not WanUpdateFilter[event.efid] or
		WanUpdateFilter[event.efid].messageid ~= event.messageid or
		WanUpdateFilter[event.efid].param0 ~= event.param0 or
		WanUpdateFilter[event.efid].param1 ~= event.param1 then
			WanUpdateFilter[event.efid] = event
			event.efid = nil	-- efid is for internal use only - don't send
			transport:send(event)
		end
	end
end

-- Handle WAN state notification from ALEOS.
-- The action runs after a delay, to prevent multiple notifications from
-- creating a flurry of DELS events (allow time to settle down first).
local WanUpdateTimer
local function dlyUpdateWanEvents(evts)
	if not WanUpdateTimer then
		WanUpdateTimer = timer.once(3,
		function() 
		-- Do work here
		updateWanEvents()
		end )
	else
		timer.rearm(WanUpdateTimer)
	end
end

-- Register for change notifications in response to changing conditions
local function registertoaleos()
	local registerIds = {}	-- table of registration IDs
	local parms = {
		msciid.STS_TIME_VALID,
		msciid.STS_DRECKON_SENSOR_CALIBRATED,
		"system.gps.fix",
		"system.gps.sat_cnt",
	}
	for k,v in pairs(vbistartparms) do
		table.insert(parms, v)
	end
	for k,v in pairs(vbiendparms) do
		table.insert(parms, v)
	end

	-- Due to a bug in ALEOS, registering all of the required notifications overflows
	-- the CSM queue. The work-around here slows down the registration process so that
	-- this does not occur.
--	devicetree.register(parms, msciidtoevent)
	for i=1, #parms do table.insert(registerIds, devicetree.register(parms[i], msciidtoevent)) sched.wait(0.2) end

--	devicetree.register(wanstate.msciids, dlyUpdateWanEvents)
	for i=1, #wanstate.msciids do table.insert(registerIds, devicetree.register(wanstate.msciids[i], dlyUpdateWanEvents)) sched.wait(0.2) end

	return registerIds
end

local function checkstartupdown()
	-- Only send boot event one per boot (in case app re-starts).
	-- If flag file (in ramdisk, gets cleared at boot) is present,
	-- startup has already been processed (AMMER is restarting), so bypass.
	-- If flag file not present this is first time since boot, so create it and process.
	local f = io.open(bootflagfile, "r")
	if not f then
		f = io.open(bootflagfile, "w")
		io.close(f)

		local uncleanshutdown = persist.load("lastheartbeat")

		if uncleanshutdown then
			generateevent("uncleanshutdown")
		else
			generateevent("shutdown")
		end
		persist.save("lastshutdown", nil)

		generateevent("startup")
		generateevent("bootignitionstate")
		generateevent("dreckon_sensor_calibration")
	else
		io.close(f)
	end

    generateevent("appinfo")
    generateevent("caninfo")
end

------------------------------------------------------------------
-- format_telemetry_msg - package telemetry data into DELS event
-- 
-- Inputs:
--	data		- table of parameter/value pairs
--
-- Outputs:
--
-- Returns:
--
-- Notes
--	Passed to temeletry constructor, this recieves telemetry data
--	and reformats/packages into DELS events
------------------------------------------------------------------
local function format_telemetry_msg(data)
	for k,v in pairs(data) do
		if k == "ignition_state" then
			-- send ignition state telemetry event as a DELS event.
			local evt = {}
			evt.messageid = events.ignitionstate.messageid
			evt.param0 = v
			evt.param1 = tonumber(devicetree.get(IGNITION_SENSE_TIME))
			transport:send(evt)
		elseif DELSparams[k] ~= nil then
			local extra = "MOI="..DEVICE_ID..",MessageString="..tostring(DELSparams[k].id).." "..DELSparams[k].format.." "..tostring(v)
			local evt = {}
			log(APP_NAME, "DETAIL", 'DELS telemetry data: "'..extra..'"')
			evt.messageid = 0
			evt.param0 = 1002
			evt.extra = extra
			transport:send(evt)
		else
			if k == "error" and v == "OBDII disconnected" then
				-- Streamer disconnect sent as battery voltage message
				transport:send({messageid=0, param0=1002, extra="MOI="..DEVICE_ID..",MessageString=83 S DISCONNECTED"})
			elseif k == "info" and v == "OBDII connected" then
				-- Streamer connect sent as battery voltage message
				transport:send({messageid=0, param0=1002, extra="MOI="..DEVICE_ID..",MessageString=83 S CONNECTED"})
			elseif k == "error" then
				transport:inform("error", v)
			elseif k == "warning" then
				transport:inform("warning", v)
			elseif k == "info" then
				transport:inform("info", v)
			else
				log(APP_NAME, "INFO", "Unhandled telemetry event %s = %s", k, v)
			end
			-- log(APP_NAME, "INFO", "Raw telemetry data: "..tostring(k).."="..tostring(v))
		end
	end
end

-- ----------------------------------------------------------------------------
-- PROCESSES
-- ----------------------------------------------------------------------------
------------------------------------------------------------------
-- main - Main App Process 
-- 
-- Inputs:
--
-- Outputs:
--
-- Returns:
--
-- Notes
------------------------------------------------------------------
local function main()
    assert(devicetree.init())
    assert(airvantage.init())
	
	local serial_port

	log(APP_NAME, "INFO", APP_NAME.." Started")
	--print("Executing from: "..os.getenv("PWD"))

	-- Stop ALEOS from spamming the log with M3DA junk (AMMER gets blamed,
	-- but it's really an AAF issue). Don't override an existing setting.
	-- Each of these modules spams the log. Lua console command tt.logfix sets
	-- another level (eg tt.logfix "INFO"). Log level nil removes settings.
	-- Note: boot is required for this to take effect.
	local waitboot = false
	if not devicetree.get("config.log.moduleslevel.M3DA-SESSION") then
		log(APP_NAME, "INFO", "Setting M3DA-SESSION log level to NONE")
		devicetree.set("config.log.moduleslevel.M3DA-SESSION", "NONE")
		waitboot=true
	end
	if not devicetree.get("config.log.moduleslevel.SRVCON") then
		log(APP_NAME, "INFO", "Setting SRVCON log level to NONE")
		devicetree.set("config.log.moduleslevel.SRVCON", "NONE")
		waitboot=true
	end
	if not devicetree.get("config.log.moduleslevel.DATAMGR") then
		log(APP_NAME, "INFO", "Setting DATAMGR log level to NONE")
		devicetree.set("config.log.moduleslevel.DATAMGR", "NONE")
		waitboot=true
	end

	if waitboot then
		-- Set up just enough to be able to get a message out to AMM
		transport = _m3da(APP_NAME, "eventrpt", "event")	-- note no wanstate param
		transport:run()
		transport:inform("warning", "AMMER Initial install - must reboot GW to activate")

		-- debugging stuff
		tt = {}
		tt.devicetree = devicetree
		tt.wanstate = wanstate
		tt.transport = transport

		-- abort main - other routines will keep app active
		return
	end

	DEVICE_ID = devicetree.get("system.hw_info.global_id")

	-- Determing which line to monitor for ignition input
	-- If MP70 telemetry enabled and using I/O for ignition
	-- then monitor I/O, else monitor ignition.
	-- (Non-4.9.0 or non-MP70 won't get these options enabled)
	IGNITION_SENSE_LINE = "system.io.ignition_in"
	IGNITION_SENSE_TIME = msciid.CFG_LOW_PWR_IGNITION_TIMER
	if devicetree.get(msciid.CFG_TELEM_CANBUS_ENABLED) == "1" and
	   devicetree.get(msciid.CFG_TELEM_IGN_SOURCE) == "2" then
		IGNITION_SENSE_LINE = "system.io.gpio.1"
		IGNITION_SENSE_TIME = msciid.CFG_LOW_PWR_IO_LINE_TIMER
	end

	transport = _m3da(APP_NAME, "eventrpt", "event")
	BnB = _BnB(APP_NAME)

	local bbem = _bbem("127.0.0.1", 0xbb5e)
	if BnB:connect(bbem) then
		log(APP_NAME, "INFO", "Using BBSE Streamer interface")
	else
		bbem:close()
		bbem=nil
		serial_port = serial.open("/dev/ttyS0", {baudRate=115200,flowControl="none"})
		BnB:connect(serial_port)
		log(APP_NAME, "INFO", "Using serial Streamer interface")
	end

	telemetry = _telemetry(APP_NAME, BnB, format_telemetry_msg)
	
	if not transport or not BnB or not telemetry then
		log(APP_NAME, "ERROR", "Cannot initialize")
		return
	end

	telemetry.options = persist.load(CONFIG_ASSET_NAME)
	if not telemetry.options then
		telemetry.options = {}
	end
	BnB:get_default_options(telemetry.options)
	-- load and activate system options
	for soptname, soptval in pairs(default_system_options) do
		if not telemetry.options[soptname] then
			telemetry.options[soptname] = soptval.value
		end
		if soptval.handler then
			soptval.handler(soptname, telemetry.options[soptname])
		end
	end

	-- Set up the asset for configuration
	local st,err = airvantage.init()
	if not st then
		log(APP_NAME, "ERROR", "Cannot initialize av lib %s", err)
		return
	end

	configasset = airvantage.newAsset(CONFIG_ASSET_NAME)
	configasset.tree.downlink = recieve_config
	configasset.tree.uplink = request_config
	configasset:start()

    transport:run()

	-- trap signal to get app shutdown notification
	psignal.signal("SIGTERM",
		function()
			local now = os.time()
			log(APP_NAME, "WARNING", APP_NAME.." SIGTERM Received ...")
			-- remove unclean shutdown marker lastheartbeat
			persist.save("lastheartbeat", nil)
			
			-- save shutdown time and reason. On first execution will generate shutdown event
			-- using this timestamp and code. Note that ALEOS sometimes sends this signal
			-- more than once, so ignore all but first (reset at boot).
			if not persist.load("lastshutdown") then
				persist.save("lastshutdown", {time = now, reason = 0} )
			end
			
			-- signal rest of app to stop
			sched.signal(APP_NAME, "EXIT")
			log(APP_NAME, "WARNING", APP_NAME.." Exiting ...")
		end
	)

	-- generate startup/shutdown events
	checkstartupdown()
	
	local last_heartbeat = 0

	-- Set up WAN state module before first use, then trigger for initial update
	wanstate.init(devicetree,
		function(message) log(APP_NAME, "DEBUG", message) end
	)
	updateWanEvents()
	
	-- Register for various operational events
    local registerIds = registertoaleos()
	
	-- For debug access via telnet console
	if DEBUG then
		tt = {
			devicetree = devicetree,
			persist = persist,
			getevent = getevent,
			generateevent = generateevent,
			split = split,
			transport = transport,
			telemetry = telemetry,
			wanstate = wanstate,
			updateWanEvents= updateWanEvents,
			dlyUpdateWanEvents = dlyUpdateWanEvents,
			WanUpdateFilter = WanUpdateFilter,
			events = events,
			configasset = configasset,
			vbistartparms = vbistartparms,
			vbiendparms = vbiendparms,
			default_system_options = default_system_options,
			recieve_config = recieve_config,
			request_config = request_config,
			msciidtoevent = msciidtoevent,
			msciid = msciid,
			registerIds = registerIds,
			-- list of options/values (eg tt.setopt {heartbeat="enable=1, min=120", sys_log_level="DETAIL"} )
			setopt = function(opts) recieve_config(configasset, opts, "uplink") end,
			getopt = function() return get_config(configasset, "", "downlink") end,
			logfix = function(level)	-- set squelched logs to different level
				devicetree.set("config.log.moduleslevel.M3DA-SESSION", level)
				devicetree.set("config.log.moduleslevel.SRVCON", level)
				devicetree.set("config.log.moduleslevel.DATAMGR", level)
				print("Don't forget to reboot for changes to take effect.")
			end,
		}
	end

	telemetry:start()

		-- Main loop - sit and poll gps and stuff
	while true do
		
		-- The following are polled periodically, but are sent
		-- according to telemetry filter rules
		generateevent("heartbeat")
		generateevent("wanlinktxbytes")
		generateevent("wanlinkrxbytes")
		-- poll for hardware-based ignition state if no streamer present
		if not telemetry:isPresent() then
			generateevent("ignitionstate")
		end
		if devicetree.get('system.gps.fix') ~= 0 then
			generateevent("gpslocation")
			generateevent("gpsspeed")
		end

		local sig = sched.wait(APP_NAME, {"EXIT", 1})
		if sig == "EXIT" then
			break
		end
	end

	print("Exiting " .. APP_NAME .. " cleaning up ...")

	-- Execute orderly shutdown
	telemetry:stop()
	if bbem then
		bbem:close()
	else
		serial_port:close()
	end
	transport:stop()
	configasset:close()
	os.exit()
end

------------------------------------------------------------------
-- main - Debug shell 
-- 
-- Inputs:
--
-- Outputs:
--
-- Returns:
--
-- Notes
--	Start a telnet server on port 1234
--	Once this program is started, you can start a Lua VM through telnet
--	using the following command: telnet localhost 1234
------------------------------------------------------------------
if DEBUG then
	local function run_server()
	local shell = require 'shell.telnet'
	
	shell.init {
		address     = '127.0.0.1',
		port        = 1234,
		editmode    = "edit",
		historysize = 100 }
	end

	sched.run(run_server)
end

sched.run(main)
sched.loop()
